(* ========================================================================= *)
(*                                                                           *)
(*        		   Application.ml				     *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2015                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <beillahi@encs.concordia.ca>,				     *)
(*		<muh_sidd@ece.concordia.ca>                                  *)
(*       	                                 			     *)
(*                                                                           *)
(* Last update: January 26, 2015                                             *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)



needs "Stability_Resonance.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*-------------------------------------------------------------------*)
(*---------------------Incohrent DCDR circuit-----------------------*)
(*-------------------------------------------------------------------*)

let Incoherent_DCDR = prove (`SFG_Main ([1,Cx(&1)-k1,3;1,k1,4;2,k1,3;2,Cx(&1)-k1,4;
3,T1,6;4,T2,5;5,k2,8;5,Cx(&1)-k2,7;6,k2,7;6,Cx(&1)-k2,8;7,T3,2], 8, 8,1) = 
((Cx(&1)-k1)*(Cx(&1)-k2)*T1 + k1*k2*T2 - 
(Cx(&1)-Cx(&2)*k1)*(Cx(&1)-Cx(&2)*k2)*T1*T2*T3) 
/ (Cx(&1) - k1*k2*T1*T3 - (Cx(&1)-k1)*(Cx(&1)-k2)*T2*T3)`,
SFG_TAC
THEN COMPLEX_DIV_TAC);; 

(*----------------------------Stability-------------------------------------*)
(*-----------------Incoherent DCDR definition------------------*)
let DCDR = 
	define ` DCDR T1 T2 T3 k1 k2 = 
	([1,Cx(&1)-k1,3;1,k1,4;2,k1,3;2,Cx(&1)-k1,4;
3,T1,6;4,T2,5;5,k2,8;5,Cx(&1)-k2,7;6,k2,7;6,Cx(&1)-k2,8;7,T3,2], 8, 8,1)`;;

(*--------------------------------------------------------------------*)
let DENOM_OF_DCDR = 
      prove (` !T1 T2 T3 k1 k2. DENOMINATOR (DCDR T1 T2 T3 k1 k2) =  (Cx (&1) -
          k1 * k2 * T1 * T3 -
          (Cx (&1) - k1) * (Cx (&1) - k2) * T2 * T3)`,
REWRITE_TAC [DCDR] THEN 
DENOMINATOR_TAC
THEN COMPLEX_DIV_TAC);;
(*-----------------------------------------------------------*)
let VSUM_22 = 
	prove (`!t. vsum(0..2) t = t(0) + t(1) + t(2)`, 
	REWRITE_TAC[num_CONV `1`; num_CONV `2`;
	VSUM_CLAUSES_NUMSEG] THEN 
	REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; VECTOR_ADD_ASSOC]);;

(*----------------Incoherent DCDR stability conditions------------------*)
let ROOT_BASIC = 
	prove ( `!z a. z = csqrt a /\ ~(a = Cx (&0)) ==> 
	Cx (&1) - a * inv z * inv z = Cx (&0)`, 
	REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REWRITE_TAC[GSYM COMPLEX_POW_2;COMPLEX_POW_INV;CSQRT] 
	THEN REPEAT(POP_ASSUM MP_TAC) 
	THEN TRY(CONV_TAC COMPLEX_FIELD));;
	
let aid0 = 
	prove (  `!(x:real). (--x) pow 2 = x pow 2 `, 
	CONV_TAC REAL_FIELD);;

let norm_neg = 
	prove (  `!(x:complex).  norm(--x) = norm(x) `, 
	REWRITE_TAC [complex_norm;RE_NEG;IM_NEG;aid0]);;
let aid2 = 
	prove ( `norm (--csqrt (k1 * k2*G1*G3 + 
	(Cx (&1) - k1) * (Cx (&1) - k2)*G2*G3)) = 
	norm (csqrt (k1 * k2*G1*G3 + (Cx (&1) - k1) * 
	(Cx (&1) - k2)*G2*G3))`, 
	REWRITE_TAC [norm_neg]);;
	
let thm_aid52 = prove( `{z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} SUBSET {z |  z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)}`, SIMP_TAC [SUBSET]
THEN SIMP_TAC [IN_ELIM_THM]);;


 let thm_aid53 = prove( `~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0)) 
 ==> (csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) IN {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} ) /\ (--csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) IN {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} ) ==> {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)} SUBSET {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} `, SET_TAC [ ]);;


let aid_thm54 = prove(`~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0)) 
==> (csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) IN {z | ~(z = Cx (&0))
      /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
Cx (&0)}) /\ (--csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) IN {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)})   `,
REWRITE_TAC[IN_ELIM_THM] THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT]
THEN SIMP_TAC [COMPLEX_POW_2] THEN SIMP_TAC [COMPLEX_MUL_LNEG;COMPLEX_MUL_RNEG]
THEN SIMP_TAC [COMPLEX_NEG_NEG] THEN SIMP_TAC [GSYM COMPLEX_POW_2]
THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT] THEN CONV_TAC COMPLEX_FIELD);;


let aid_thm55 = prove( `~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0)) 
==> {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)} SUBSET {z | ~(z = Cx (&0)) /\ z pow 2 -
    (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)}`,
STRIP_TAC THEN MP_TAC aid_thm54 THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC [thm_aid53]);;


let thm_aid58 = prove( ` {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)} SUBSET {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} /\ (CARD ({z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)}) <= CARD {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)}) /\ (FINITE ({z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)}) ) ==> {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} = {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)}  `,
MESON_TAC[CARD_SUBSET; CARD_SUBSET_EQ; LE_ANTISYM]);;


let thm_aid32 = prove( ` ~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0)) 
==> CARD {csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3), --
 csqrt(k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3)}  = 2 `,
SIMP_TAC[CARD_CLAUSES; FINITE_INSERT; FINITE_EMPTY; NOT_IN_EMPTY; ARITH]
THEN SIMP_TAC [ IN_SING ] THEN CONV_TAC COMPLEX_FIELD);;


let incoherent_denom_poly = prove( ` z pow 2 -
      (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = vsum (0..2) (\i. (EL i [
  --(k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3);Cx(&0) ; Cx (&1)]) * 
  ( z) pow i)`,
REWRITE_TAC[ONE;TWO;VSUM_22;EL;HD;TL;complex_pow] THEN 
CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD);;


 let thm_aid24 = prove( `2 IN 0..2`,
SIMP_TAC[GSYM NUMSEG_LE] THEN REWRITE_TAC[IN_ELIM_THM] 
THEN CONV_TAC NUM_REDUCE_CONV );;


let aid_thm23 = prove( ` ~( EL 2
           [--(k1 * k2 * G1 * G3 +
               (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3); 
           Cx
           (&0); Cx (&1)] = Cx(&0) )`, SIMP_TAC [EL;HD;TL;ONE;TWO]
		   THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD );;


 let thm_aid25 = prove( `~(2 IN 0..2
 ==> EL 2
 [--(k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3); 
           Cx
           (&0); Cx (&1)] =
           Cx (&0))`,
SIMP_TAC [aid_thm23;thm_aid24]);;


let thm_goal = prove( `~(!i. i IN 0..2 ==>  (EL i
           [--(k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3); 
           Cx
           (&0); Cx (&1)]) = Cx (&0))`,
 REWRITE_TAC[ NOT_FORALL_THM] THEN MESON_TAC [thm_aid25]);;


let goal_thm26 = prove( ` CARD {z| (z pow 2 - (k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3)) = Cx(&0)} <= 2 
/\ FINITE {z| (z pow 2 - (k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3)) = Cx(&0)}`,
MP_TAC thm_goal THEN SIMP_TAC [incoherent_denom_poly]
THEN SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND]);;


let incoherent_finite_set_poles = 
prove( `  FINITE {z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} `,
MP_TAC thm_aid52 THEN MP_TAC goal_thm26
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC []
THEN MESON_TAC [ FINITE_SUBSET]);;


let thm_aid56 = prove( `  CARD ({z | ~(z = Cx (&0))
               /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
Cx (&0)}) <= CARD {z| (z pow 2 - (k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3)) = Cx(&0)}` ,
MP_TAC thm_aid52 THEN MP_TAC goal_thm26
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC []
THEN MESON_TAC [CARD_SUBSET  ]);;



let thm_aid33 = prove(` ( ~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0))) 
==> (CARD ({z| (z pow 2 - (k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3)) = Cx(&0)}) 
<= CARD {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)}) `,
 SIMP_TAC [goal_thm26;thm_aid32]);;

 
let incoherent_card_set_poles = 
prove( ` ( ~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0))) 
==> (CARD ({z | ~(z = Cx (&0)) /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)}) <= CARD {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)}) `,
STRIP_TAC THEN MP_TAC thm_aid33 THEN MP_TAC thm_aid56
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN ASM_ARITH_TAC);;


let incoherent_set_poles = 
prove( `  ( ~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0))) 
==> {z | ~(z = Cx (&0)) /\ z pow 2 -
                   (k1 * k2 * G1 * G3 +
                    (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) =
                   Cx (&0)} = {csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3),
	--csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*(Cx(&1)-k2)*G2*G3)}  `,
STRIP_TAC THEN MP_TAC aid_thm55 THEN MP_TAC incoherent_finite_set_poles
THEN MP_TAC incoherent_card_set_poles THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC [thm_aid58]);;
 
 
let aid_thm50 =prove( `a IN {b, c} <=> (a = c) \/ (a=b)`,
SET_TAC []);;


let DENOMINATOR_OF_DCDR = prove( ` ~(z = Cx (&0))
 ==>  DENOMINATOR 
(DCDR (G1/z) (G2/z) (G3/z) k1 k2) =  
(z pow 2 - (k1*k2*G1*G3+(Cx (&1)-k1)*(Cx (&1)-k2)*G2*G3))/(z pow 2)`,
SIMP_TAC[DENOM_OF_DCDR] THEN CONV_TAC COMPLEX_FIELD);;


let aid_thm51 = prove (`~(z = Cx (&0)) /\ P <=>  ~(z = Cx (&0)) /\
     (~(z = Cx (&0))
      ==> P)`, MESON_TAC []);;

let aid_thm61 = prove (` ~(z = Cx (&0)) /\
     (P \/
      z = Cx (&0)) <=>
     ~(z = Cx (&0)) /\
     P`, MESON_TAC []);;


let stability= prove(` (norm (csqrt (k1*k2*G1*G3+(Cx(&1)-k1)*
	(Cx(&1)-k2)*G2*G3)) <= &1) /\ 
	~(csqrt (k1 * k2 * G1 * G3 + (Cx (&1) - k1) * (Cx (&1) - k2) * G2 * G3) = Cx(&0)) 
	==> (is_stable_psp (\z. DCDR (G1/z) (G2/z) 
		(G3/z) k1 k2) z )`,
SIMP_TAC [is_stable_psp;poles_in_unit_circle;
	are_valid_poles;in_unit_circle;
	is_valid_pole;set_of_poles;DENOMINATOR_OF_DCDR] THEN
 ONCE_REWRITE_TAC [aid_thm51] THEN SIMP_TAC [DENOMINATOR_OF_DCDR] 
THEN ONCE_REWRITE_TAC [GSYM aid_thm51] THEN SIMP_TAC [COMPLEX_DIV_EQ_0]
THEN SIMP_TAC [ (COMPLEX_FIELD ` (z pow 2 = Cx(&0)) <=> (z = Cx(&0))`)]
THEN ONCE_REWRITE_TAC [aid_thm61] THEN SIMP_TAC [incoherent_set_poles]
THEN SIMP_TAC [aid_thm50] THEN REPEAT STRIP_TAC
THEN ASM_SIMP_TAC [] THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC[aid2] THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC[]);;

	 
	
(*----------------Numerator of Incoherent DCDR-------------------*)
let NUM_OF_DCDR = prove (` !T1 T2 T3 k1 k2. NUMERATOR (DCDR T1 T2 T3 k1 k2) =  
	((Cx(&1)-k1)*(Cx(&1)-k2)*T1 + k1*k2*T2 - 
	(Cx(&1)-Cx(&2)*k1)*(Cx(&1)-Cx(&2)*k2)*T1*T2*T3)`,
REWRITE_TAC [DCDR] THEN 
NUMERATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;

(*-----------------------------------------------------------*)
let VSUM_33 = 
	prove (`!t. vsum(0..3) t = t(0) + t(1) + t(2) + t(3)`, 
	REWRITE_TAC[num_CONV `1`; num_CONV `2`;  num_CONV `3`;  
	VSUM_CLAUSES_NUMSEG] THEN 
	REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; VECTOR_ADD_ASSOC]);;

(*-------------Incoherent DCDR resonance conditions-----------------*)
let ZERO_BASIC =
     prove ( `!z a b. z = csqrt (b/a) /\ ~(b = Cx (&0)) /\ ~(a = Cx (&0)) 
     ==> a * inv z - b * inv z * inv z * inv z = Cx (&0)`,
      REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
      REWRITE_TAC[GSYM COMPLEX_POW_2;COMPLEX_POW_INV;CSQRT] THEN 
      REPEAT(POP_ASSUM MP_TAC) THEN TRY(CONV_TAC COMPLEX_FIELD));;


let NUMERATOR_OF_DCDR = prove( `~(z = Cx (&0))
 ==> NUMERATOR (DCDR (G1/z) (G2/z) (G3/z) k1 k2) =   ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3))/(z pow 3)`,
SIMP_TAC[NUM_OF_DCDR] THEN CONV_TAC COMPLEX_FIELD);;



let incoherent_num_poly = prove( ` ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = vsum (0..2) (\i. (EL i [
  --((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3);Cx(&0) ; (k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)]) * 
  ( z) pow i)`,
 REWRITE_TAC[ONE;TWO;VSUM_22;EL;HD;TL;complex_pow] 
THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD);;


 let thm_aid24 = prove( `2 IN 0..2`,
SIMP_TAC[GSYM NUMSEG_LE] THEN 
REWRITE_TAC[IN_ELIM_THM]  THEN
CONV_TAC NUM_REDUCE_CONV );;
 

let aid_thm72 = prove( ` (~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) ==> ~( EL 2
[--((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3);Cx(&0) ; (k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)] = Cx(&0) )`, 
SIMP_TAC [EL;HD;TL;ONE;TWO] THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD );;


let thm_aid73 = prove( `(~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
==> ~(2 IN 0..2
       ==> EL 2
[--((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3);Cx(&0) ; (k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)] =
           Cx (&0))`,
SIMP_TAC [aid_thm72;thm_aid24]);;





let thm_aid74 = prove( `(~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
==> ~(!i. i IN 0..2 ==>  (EL i
[--((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3);Cx(&0) ; (k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)]) = Cx (&0))`,
REWRITE_TAC[ NOT_FORALL_THM] THEN MESON_TAC [thm_aid73]);;






 let thm_aid75 =  prove( `~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) 
	==> (csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) IN {z| ~(z = Cx (&0))
	/\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	/\ (--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) IN {z| ~(z = Cx (&0))
	/\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)})   `,
REWRITE_TAC[IN_ELIM_THM] THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT]
THEN SIMP_TAC [COMPLEX_POW_2] THEN SIMP_TAC [COMPLEX_MUL_LNEG;COMPLEX_MUL_RNEG]
THEN SIMP_TAC [COMPLEX_NEG_NEG] THEN SIMP_TAC [GSYM COMPLEX_POW_2]
THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT]
THEN SIMP_TAC [ (COMPLEX_FIELD ` ~(csqrt
   (((Cx (&1) - Cx (&2) * k1) * (Cx (&1) - Cx (&2) * k2) * G1 * G2 * G3) /
    ((Cx (&1) - k1) * (Cx (&1) - k2) * G1 + k1 * k2 * G2)) =
   Cx (&0)) ==> ~(--csqrt
   (((Cx (&1) - Cx (&2) * k1) * (Cx (&1) - Cx (&2) * k2) * G1 * G2 * G3) /
    ((Cx (&1) - k1) * (Cx (&1) - k2) * G1 + k1 * k2 * G2)) =
   Cx (&0))`)]
THEN REWRITE_TAC[CSQRT_EQ_0;COMPLEX_DIV_EQ_0;DE_MORGAN_THM] THEN CONV_TAC COMPLEX_FIELD);;


 let thm_aid76 = prove( `~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) 
	==> (csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) IN {z| ~(z = Cx (&0)) 
	/\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	/\ (--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) IN {z| ~(z = Cx (&0)) 
	/\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	==> {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))} SUBSET {z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} `,
 SET_TAC [ ]);;


let aid_thm77 = prove( `~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) 
	==> {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))} SUBSET {z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}`,
STRIP_TAC THEN MP_TAC thm_aid75 THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC [thm_aid76]);;


let thm_aid78 = prove( `{z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} 
	SUBSET {z | ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}`,
SIMP_TAC [SUBSET] THEN SIMP_TAC [IN_ELIM_THM]);;


let goal_thm79 = prove(` (~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
==> CARD {z| ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} <= 2 
	/\ FINITE {z| ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}`,
STRIP_TAC THEN MP_TAC thm_aid74 THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC [incoherent_num_poly] THEN SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND]);;


let thm_aid80 = prove(`  (~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
==> CARD ({z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	<= CARD {z | ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}` ,
STRIP_TAC THEN MP_TAC thm_aid78 THEN MP_TAC goal_thm79
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [CARD_SUBSET  ]);;


let incoherent_finite_set_zeros = 
prove( `  (~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
==> FINITE {z | ~(z = Cx (&0))
   /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}`,
STRIP_TAC THEN MP_TAC thm_aid78 THEN MP_TAC goal_thm79
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [ FINITE_SUBSET]);;







let thm_aid82 = prove( ` {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))} SUBSET {z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} /\ (CARD ({z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	<= CARD {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}) /\ (FINITE ({z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) ) ==> {z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} = 
	{csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}  `,
MESON_TAC[CARD_SUBSET; CARD_SUBSET_EQ; LE_ANTISYM]);;




let thm_aid83 = prove( ` ~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0))
	==> CARD {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}  = 2 `,
SIMP_TAC[CARD_CLAUSES; FINITE_INSERT; FINITE_EMPTY; NOT_IN_EMPTY; ARITH]
THEN SIMP_TAC [ IN_SING ] THEN CONV_TAC COMPLEX_FIELD);;



let thm_aid84 = prove( ` ~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) /\ 
	(~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) 
	==> (CARD ({z| ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	<= CARD {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}) `,
SIMP_TAC [goal_thm79;thm_aid83]);;


let incoherent_card_set_zeros= 
prove(` ~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) /\ 
	(~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0)))
	==> (CARD ({z | ~(z = Cx (&0))
 /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)}) 
	<= CARD {csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}) `,
STRIP_TAC THEN MP_TAC thm_aid84 THEN MP_TAC thm_aid80
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN ASM_ARITH_TAC);;


let thm_aid86= 
prove( `  ~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) 
	==>  (~(((Cx (&1) - k1) * (Cx (&1) - k2)*G1 + k1* k2*G2 ) = Cx(&0))) `,
SIMP_TAC[CSQRT_EQ_0;COMPLEX_DIV_EQ_0;DE_MORGAN_THM]);;


let incoherent_set_zeros = 
prove( `  ~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) ==> {z | ~(z = Cx (&0))
               /\ ((k1*k2*G2+(Cx (&1)-k1)*(Cx (&1)-k2)*G1)* z pow 2
	-((Cx (&1)-(Cx (&2))*k1)*(Cx (&1)-(Cx (&2))*k2)*G1*G2*G3)) = Cx(&0)} = 
	{csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )),
	--csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))}  `,
STRIP_TAC THEN MP_TAC thm_aid86 THEN MP_TAC incoherent_card_set_zeros
THEN MP_TAC incoherent_finite_set_zeros THEN MP_TAC aid_thm77
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [thm_aid82]);;

 
let zero_aid0 = 
	prove ( `norm 
	(--csqrt (((Cx (&1) - (Cx (&2))*k1) * 
	(Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))) = 
	norm (csqrt (((Cx (&1) - (Cx (&2))*k1) * 
	(Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
	((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )))`, 
	REWRITE_TAC [norm_neg]);;


let resonance= prove(`(norm 
(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 ))) <= &1) /\
~(csqrt (((Cx (&1) - (Cx (&2))*k1) * (Cx (&1) - (Cx (&2))*k2)*G1*G2*G3)/
((Cx (&1) - k1) * (Cx (&1) - k2)*G1+ k1 * k2*G2 )) = Cx(&0)) 
==> is_resonant_psp (\z. DCDR (G1/z) (G2/z) (G3/z) k1 k2) z `,
REWRITE_TAC [is_resonant_psp;poles_in_unit_circle;
	are_valid_zeros;in_unit_circle;
	is_valid_zero;set_of_zeros;NUMERATOR_OF_DCDR] THEN
ONCE_REWRITE_TAC [aid_thm51] THEN SIMP_TAC [NUMERATOR_OF_DCDR] 
THEN ONCE_REWRITE_TAC [GSYM aid_thm51] THEN SIMP_TAC [COMPLEX_DIV_EQ_0]
THEN SIMP_TAC [ (COMPLEX_FIELD ` (z pow 3 = Cx(&0)) <=> (z = Cx(&0))`)]
THEN ONCE_REWRITE_TAC [aid_thm61] THEN SIMP_TAC [incoherent_set_zeros]
THEN SIMP_TAC [aid_thm50] THEN REPEAT STRIP_TAC
THEN ASM_SIMP_TAC [] THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC[ zero_aid0] THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC[]);;


(*------------------------------------------------------------------*)
(*---------------------Coherent DCDR---------------------------------*)
(*-------------------------------------------------------------------*)

let help1 = 
	prove ( `!(k1:real) (k2:real) (T2:complex) (T3:complex).
       ((&0<=(&1-k1)) /\ (&0<=(&1-k2)) /\ (&0<=k1) /\ (&0<=k2)) 
	==> (Cx(sqrt (&1 - k1)) * T2 * Cx(sqrt (&1 - k2)) * T3 = 
		(Cx(sqrt (&1 - k1)) * Cx(sqrt (&1 - k2))) * T2 * T3) /\ 
		(Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - Cx (&1 - k1) * 
		Cx (&1 - k2) * T1 * T2 * T3 +
		--Cx (&1 - k1) * Cx k2 * T1 * T2 * T3 +
		--Cx (sqrt (k1 * k2)) * T2 - Cx k1 * Cx k2 * T1 * T2 * T3 +
		--Cx k1 * Cx (&1 - k2) * T1 * T2 * T3 +
		Cx (&0) = Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - 
		Cx (sqrt (k1 * k2)) * T2 - 
		((Cx (&1 - k1) * Cx (&1 - k2)) + (Cx (&1 - k1) * Cx k2) + 
		(Cx k1 * Cx k2) + (Cx k1 * Cx (&1 - k2))) 
	* T1 * T2 * T3) /\ ((Cx (sqrt (&1 - k1)) * T1 * Cx (sqrt (&1 - k2))) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) = 
	(Cx (sqrt (&1 - k1)) * Cx (sqrt (&1 - k2))) * T1 - 
	(Cx (sqrt (&1 - k1)) * Cx (sqrt (&1 - k1))) * 
	(Cx (sqrt (&1 - k2)) * Cx (sqrt (&1 - k2)))* T1 * T2 * T3) /\ 
	((Cx (sqrt (&1 - k1)) *
           T1 *
           (--ii * Cx (sqrt k2)) *
           T3 *
           Cx (sqrt (&1 - k1)) *
           T2 *
           --ii *
           Cx (sqrt k2)) *
          (Cx (&1) + Cx (&0)) = (Cx (sqrt (&1 - k1)) *
           Cx (sqrt (&1 - k1))) * --(Cx (sqrt k2) * 
		   Cx (sqrt k2))*T1*T2*T3) /\ (((--ii * Cx (sqrt k1)) *
           T2 *
           Cx (sqrt (&1 - k2)) *
           T3 *
           (--ii * Cx (sqrt k1)) *
           T1 *
           Cx (sqrt (&1 - k2))) *
          (Cx (&1) + Cx (&0)) = -- (Cx (sqrt k1)*Cx (sqrt k1))*
		  (Cx (sqrt (&1 - k2)) * Cx (sqrt (&1 - k2)))*T1*T2*T3) /\ 
		  (((--ii * Cx (sqrt k1)) * T2 * --ii * Cx (sqrt k2)) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) = 
		   --(Cx (sqrt k1)*Cx (sqrt k2))*T2-(Cx (sqrt k1)*Cx (sqrt k1))*
		   (Cx (sqrt k2)*Cx (sqrt k2))*T1*T2*T3) /\ ((Cx (&1) +
          Cx (&0) +
          Cx (&0) +
          (Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) -
          (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) = 
		  Cx (&1) + (Cx (sqrt k1)*Cx (sqrt k2))*T1*T3-
		  (Cx (sqrt (&1 - k1))*Cx (sqrt (&1 - k2)))*T2*T3)`,
	 REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN CONV_TAC COMPLEX_FIELD);;

let help2 = 
	prove ( `!(k1:real) (k2:real).
       (&1 - k1) * (&1 - k2) + (&1 - k1) * k2 + k1 * k2 + k1 * (&1 - k2) = &1`, 
	   CONV_TAC REAL_FIELD);;


let help3 = 
	prove ( `!(k1:real) (k2:real) (T2:complex) (T3:complex).
       ((&0<=k1) /\ (&0<=(&1-k2)) /\ (&0<=k2) /\ (&0<=(&1-k1))) 
	==> (Cx(sqrt (&1 - k1)) * T2 * Cx(sqrt (&1 - k2)) * T3 = 
	Cx(sqrt ((&1 - k1) * (&1 - k2))) * T2 * T3) /\ (((--ii * Cx (sqrt k1)) *
           T2 *
           Cx (sqrt (&1 - k2)) *
           T3 *
           (--ii * Cx (sqrt k1)) *
           T1 *
           Cx (sqrt (&1 - k2))) *
          (Cx (&1) + Cx (&0)) = -- (Cx k1)*(Cx (&1 - k2))*T1*T2*T3) /\ 
		  ( (Cx (sqrt (&1 - k1)) * T1 * Cx (sqrt (&1 - k2))) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) = 
	Cx(sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - 
	Cx(&1 - k1) * Cx(&1 - k2)* T1 * T2 * T3) /\ ((Cx (sqrt (&1 - k1)) *
           T1 *
           (--ii * Cx (sqrt k2)) *
           T3 *
           Cx (sqrt (&1 - k1)) *
           T2 *
           --ii *
           Cx (sqrt k2)) *
          (Cx (&1) + Cx (&0)) = -- Cx (&1 - k1) * Cx k2 * T1*T2*T3) /\ 
		  (((--ii * Cx (sqrt k1)) * T2 * --ii * Cx (sqrt k2)) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) =
	   --(Cx (sqrt (k1*k2)))*T2-Cx k1*Cx k2*T1*T2*T3) /\ ((Cx (&1) +
          Cx (&0) +
          Cx (&0) +
          (Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) -
          (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) =  
         Cx (&1) + (Cx (sqrt (k1* k2)))*T1*T3-(Cx (sqrt ((&1 - k1)*(&1 - k2))))
	 *T2*T3) /\
	 ( Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - Cx (&1 - k1) * 
	 Cx (&1 - k2) * T1 * T2 * T3 +
	 --Cx (&1 - k1) * Cx k2 * T1 * T2 * T3 +
	--Cx (sqrt (k1 * k2)) * T2 - Cx k1 * Cx k2 * T1 * T2 * T3 +
	--Cx k1 * Cx (&1 - k2) * T1 * T2 * T3 +
	Cx (&0) = Cx (sqrt ((&1 - k1) * (&1 - k2))) * T1 - 
	Cx (sqrt (k1 * k2)) * T2 - T1*T2*T3)  `,
	 REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [help1] 
	THEN SIMP_TAC [GSYM CX_MUL] THEN SIMP_TAC [GSYM CX_ADD] THEN 
	SIMP_TAC [GSYM REAL_POW_2] THEN SIMP_TAC [SQRT_POW_2] THEN 
	SIMP_TAC [GSYM SQRT_MUL] THEN SIMP_TAC [help2] 
	THEN CONV_TAC COMPLEX_FIELD);;
	
let Coherent_DCDR = 
    prove (`!(k1:real) (k2:real) (T1:complex) (T2:complex) (T3:complex).
((&0<=k1)/\(&0<=k2)/\(&0<=(&1-k1))/\(&0<=(&1-k2))) ==>
SFG_Main ([1,Cx(sqrt(&1-k1)),3;1,--ii*Cx(sqrt(k1)),4;2,--ii*Cx(sqrt(k1)),3;
2,Cx(sqrt(&1-k1)),4;3,T1,6;4,T2,5;5,--ii*Cx(sqrt(k2)),8;5,Cx(sqrt(&1-k2)),7;
6,Cx(sqrt(&1-k2)),8;6,--ii*Cx(sqrt(k2)),7;7,T3,2], 8, 8,1) = 
(Cx(sqrt((&1-k1)*(&1-k2)))*T1 - Cx(sqrt(k1*k2))*T2 - T1*T2*T3) 
/ (Cx(&1) + Cx(sqrt(k1*k2))*T1*T3 - Cx(sqrt((&1-k1)*(&1-k2)))*T2*T3)`,
SFG_TAC
THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC [help3]);;
(*-------------------SFG of coherent DCDR------------------------------*)
let DCDR_coherent = 
	define `!(k1:real) (k2:real) (T1:complex) (T2:complex) (T3:complex).
	DCDR_coherent T1 T2 T3 k1 k2 = 
	([1,Cx(sqrt(&1-k1)),3;1,--ii*Cx(sqrt(k1)),4;2,--ii*Cx(sqrt(k1)),3;
2,Cx(sqrt(&1-k1)),4;3,T1,6;4,T2,5;5,--ii*Cx(sqrt(k2)),8;5,Cx(sqrt(&1-k2)),7;
6,Cx(sqrt(&1-k2)),8;6,--ii*Cx(sqrt(k2)),7;7,T3,2], 8, 8,1)`;;

(*-----------------------------Helper Functions-------------------------------*)
let help1 = 
	prove ( `!(k1:real) (k2:real) (T2:complex) (T3:complex).
       ((&0<=(&1-k1)) /\ (&0<=(&1-k2)) /\ (&0<=k1) /\ (&0<=k2)) 
	==> (Cx(sqrt (&1 - k1)) * T2 * Cx(sqrt (&1 - k2)) * T3 = 
		(Cx(sqrt (&1 - k1)) * Cx(sqrt (&1 - k2))) * T2 * T3) /\ 
		(Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - Cx (&1 - k1) * 
		Cx (&1 - k2) * T1 * T2 * T3 +
		--Cx (&1 - k1) * Cx k2 * T1 * T2 * T3 +
		--Cx (sqrt (k1 * k2)) * T2 - Cx k1 * Cx k2 * T1 * T2 * T3 +
		--Cx k1 * Cx (&1 - k2) * T1 * T2 * T3 +
		Cx (&0) = Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - 
		Cx (sqrt (k1 * k2)) * T2 - 
		((Cx (&1 - k1) * Cx (&1 - k2)) + (Cx (&1 - k1) * Cx k2) + 
		(Cx k1 * Cx k2) + (Cx k1 * Cx (&1 - k2))) 
	* T1 * T2 * T3) /\ ((Cx (sqrt (&1 - k1)) * T1 * Cx (sqrt (&1 - k2))) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) = 
	(Cx (sqrt (&1 - k1)) * Cx (sqrt (&1 - k2))) * T1 - 
	(Cx (sqrt (&1 - k1)) * Cx (sqrt (&1 - k1))) * 
	(Cx (sqrt (&1 - k2)) * Cx (sqrt (&1 - k2)))* T1 * T2 * T3) /\ 
	((Cx (sqrt (&1 - k1)) *
           T1 *
           (--ii * Cx (sqrt k2)) *
           T3 *
           Cx (sqrt (&1 - k1)) *
           T2 *
           --ii *
           Cx (sqrt k2)) *
          (Cx (&1) + Cx (&0)) = (Cx (sqrt (&1 - k1)) *
           Cx (sqrt (&1 - k1))) * --(Cx (sqrt k2) * 
		   Cx (sqrt k2))*T1*T2*T3) /\ (((--ii * Cx (sqrt k1)) *
           T2 *
           Cx (sqrt (&1 - k2)) *
           T3 *
           (--ii * Cx (sqrt k1)) *
           T1 *
           Cx (sqrt (&1 - k2))) *
          (Cx (&1) + Cx (&0)) = -- (Cx (sqrt k1)*Cx (sqrt k1))*
		  (Cx (sqrt (&1 - k2)) * Cx (sqrt (&1 - k2)))*T1*T2*T3) /\ 
		  (((--ii * Cx (sqrt k1)) * T2 * --ii * Cx (sqrt k2)) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) = 
		   --(Cx (sqrt k1)*Cx (sqrt k2))*T2-(Cx (sqrt k1)*Cx (sqrt k1))*
		   (Cx (sqrt k2)*Cx (sqrt k2))*T1*T2*T3) /\ ((Cx (&1) +
          Cx (&0) +
          Cx (&0) +
          (Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) -
          (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) = 
		  Cx (&1) + (Cx (sqrt k1)*Cx (sqrt k2))*T1*T3-
		  (Cx (sqrt (&1 - k1))*Cx (sqrt (&1 - k2)))*T2*T3)`,
	 REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN CONV_TAC COMPLEX_FIELD);;

let help2 = 
	prove ( `!(k1:real) (k2:real).
       (&1 - k1) * (&1 - k2) + (&1 - k1) * k2 + k1 * k2 + k1 * (&1 - k2) = &1`, 
	   CONV_TAC REAL_FIELD);;


let help3 = 
	prove ( `!(k1:real) (k2:real) (T2:complex) (T3:complex).
       ((&0<=k1) /\ (&0<=(&1-k2)) /\ (&0<=k2) /\ (&0<=(&1-k1))) 
	==> (Cx(sqrt (&1 - k1)) * T2 * Cx(sqrt (&1 - k2)) * T3 = 
	Cx(sqrt ((&1 - k1) * (&1 - k2))) * T2 * T3) /\ (((--ii * Cx (sqrt k1)) *
           T2 *
           Cx (sqrt (&1 - k2)) *
           T3 *
           (--ii * Cx (sqrt k1)) *
           T1 *
           Cx (sqrt (&1 - k2))) *
          (Cx (&1) + Cx (&0)) = -- (Cx k1)*(Cx (&1 - k2))*T1*T2*T3) /\ 
		  ( (Cx (sqrt (&1 - k1)) * T1 * Cx (sqrt (&1 - k2))) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) = 
	Cx(sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - 
	Cx(&1 - k1) * Cx(&1 - k2)* T1 * T2 * T3) /\ ((Cx (sqrt (&1 - k1)) *
           T1 *
           (--ii * Cx (sqrt k2)) *
           T3 *
           Cx (sqrt (&1 - k1)) *
           T2 *
           --ii *
           Cx (sqrt k2)) *
          (Cx (&1) + Cx (&0)) = -- Cx (&1 - k1) * Cx k2 * T1*T2*T3) /\ 
		  (((--ii * Cx (sqrt k1)) * T2 * --ii * Cx (sqrt k2)) *
          (Cx (&1) +
           Cx (&0) +
           Cx (&0) +
           Cx (&0) - (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) =
	   --(Cx (sqrt (k1*k2)))*T2-Cx k1*Cx k2*T1*T2*T3) /\ ((Cx (&1) +
          Cx (&0) +
          Cx (&0) +
          (Cx (&0) +
           Cx (&0) +
           Cx (&0) - Cx (sqrt (&1 - k1)) * T2 * Cx (sqrt (&1 - k2)) * T3) -
          (--ii * Cx (sqrt k1)) * T1 * (--ii * Cx (sqrt k2)) * T3) =  
         Cx (&1) + (Cx (sqrt (k1* k2)))*T1*T3-(Cx (sqrt ((&1 - k1)*(&1 - k2))))
	 *T2*T3) /\
	 ( Cx (sqrt (&1 - k1) * sqrt (&1 - k2)) * T1 - Cx (&1 - k1) * 
	 Cx (&1 - k2) * T1 * T2 * T3 +
	 --Cx (&1 - k1) * Cx k2 * T1 * T2 * T3 +
	--Cx (sqrt (k1 * k2)) * T2 - Cx k1 * Cx k2 * T1 * T2 * T3 +
	--Cx k1 * Cx (&1 - k2) * T1 * T2 * T3 +
	Cx (&0) = Cx (sqrt ((&1 - k1) * (&1 - k2))) * T1 - 
	Cx (sqrt (k1 * k2)) * T2 - T1*T2*T3)  `,
	 REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [help1] 
	THEN SIMP_TAC [GSYM CX_MUL] THEN SIMP_TAC [GSYM CX_ADD] THEN 
	SIMP_TAC [GSYM REAL_POW_2] THEN SIMP_TAC [SQRT_POW_2] THEN 
	SIMP_TAC [GSYM SQRT_MUL] THEN SIMP_TAC [help2] 
	THEN CONV_TAC COMPLEX_FIELD);;
(*------------------Denominator of coherent DCDR-----------------*)

let DENOM_OF_COHERENT_DCDR = 
prove (` !(k1:real) (k2:real) (T1:complex) (T2:complex) (T3:complex).
	((&0<=k1)/\(&0<=k2)/\(&0<=(&1-k1))/\(&0<=(&1-k2))) ==>
	 DENOMINATOR (DCDR_coherent T1 T2 T3 k1 k2) =  
	(Cx(&1) + Cx(sqrt(k1*k2))*T1*T3 - Cx(sqrt((&1-k1)*(&1-k2)))*T2*T3)`,
REWRITE_TAC [DCDR_coherent]
THEN DENOMINATOR_TAC
THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC [help3]);;


(*--------------------numerator of coherent DCDR-----------------*)

let NUM_OF_COHERENT_DCDR = 
    prove (` !(k1:real) (k2:real) (T1:complex) (T2:complex) (T3:complex).
	((&0<=k1)/\(&0<=k2)/\(&0<=(&1-k1))/\(&0<=(&1-k2))) ==>
	 NUMERATOR (DCDR_coherent T1 T2 T3 k1 k2) =  
	(Cx(sqrt((&1-k1)*(&1-k2)))*T1 - Cx(sqrt(k1*k2))*T2 - T1*T2*T3)`,
REWRITE_TAC [DCDR_coherent]
THEN NUMERATOR_TAC
THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC [help3]);;



(*-------------------------------------------------------------------------*)
(* 		Stability conditions of coherent DCDR circuit              *)
(*-------------------------------------------------------------------------*)
	

let DENOMINATOR_OF_COHERENT_DCDR = 
prove( `~(z = Cx (&0)) /\ &0 <= k1 /\ &0 <= k2 /\ &0 <= &1 - k1 /\ &0 <= &1 - k2
 ==> DENOMINATOR (DCDR_coherent (G1/z) (G2/z) (G3/z) k1 k2) =   (z pow 2
	+(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3))/(z pow 2)`,
SIMP_TAC[DENOM_OF_COHERENT_DCDR] THEN CONV_TAC COMPLEX_FIELD);;


let coherent_denom_poly = 
prove(` (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = vsum (0..2) (\i. (EL i [
  (Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3);Cx(&0) ; Cx(&1)]) * 
  ( z) pow i)`,
REWRITE_TAC[ONE;TWO;VSUM_22;EL;HD;TL;complex_pow] 
THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD);;


let aid_thm91 = prove( `  ~( EL 2
           [(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3);Cx(&0) ; Cx(&1)]= Cx(&0) )`, 
		   SIMP_TAC [EL;HD;TL;ONE;TWO] THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD );;


  let thm_aid92 = prove( `~(2 IN 0..2
       ==> EL 2
           [(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3);Cx(&0) ; Cx(&1)] =
           Cx (&0))`,
 SIMP_TAC [aid_thm91;thm_aid24]);;


 let thm_aid93 = prove( `~(!i. i IN 0..2 ==>  (EL i
  [(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3);Cx(&0) ; Cx(&1)]) = Cx (&0))`,
 REWRITE_TAC[ NOT_FORALL_THM] THEN MESON_TAC [thm_aid92]);;


let thm_aid94 = prove( `~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> (csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) IN {z| ~(z = Cx (&0)) 
/\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
/\ (--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) IN {z| ~(z = Cx (&0)) 
/\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)})   `,
REWRITE_TAC[IN_ELIM_THM] THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT] THEN
SIMP_TAC [COMPLEX_POW_2] THEN SIMP_TAC [COMPLEX_MUL_LNEG;COMPLEX_MUL_RNEG]
THEN SIMP_TAC [COMPLEX_NEG_NEG] THEN SIMP_TAC [GSYM COMPLEX_POW_2]
THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT] THEN 
SIMP_TAC [ (COMPLEX_FIELD ` ~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) =
   Cx (&0)) ==> ~(--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) =
   Cx (&0))`)] THEN REWRITE_TAC[CSQRT_EQ_0;COMPLEX_DIV_EQ_0;DE_MORGAN_THM]
 THEN CONV_TAC COMPLEX_FIELD);;


 let thm_aid95 = prove( `~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
 ==> (csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) IN {z| ~(z = Cx (&0)) 
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
 /\ (--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) IN {z| ~(z = Cx (&0)) 
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
 ==> {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)} SUBSET {z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} `,
 SET_TAC [ ]);;


let aid_thm96 =prove(`~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)} SUBSET {z | ~(z = Cx (&0))
/\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} `,
STRIP_TAC THEN MP_TAC thm_aid94 THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [thm_aid95]);;


let thm_aid97 = prove( `{z | ~(z = Cx (&0))
/\  (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} 
SUBSET {z |  (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}`,
SIMP_TAC [SUBSET] THEN SIMP_TAC [IN_ELIM_THM]);;


let goal_thm98 = 
prove( ` CARD {z| (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} <= 2 /\ 
FINITE {z| (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}`,
MP_TAC thm_aid93 THEN SIMP_TAC [coherent_denom_poly]
THEN SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND]);;


let thm_aid99 = prove( `  CARD ({z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
 <= CARD {z | (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}`,
MP_TAC thm_aid97 THEN MP_TAC goal_thm98 THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [CARD_SUBSET  ]);;


let coherent_finite_set_poles = 
prove( `  FINITE {z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}`,
 MP_TAC thm_aid97 THEN MP_TAC goal_thm98 THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [ FINITE_SUBSET]);;


let thm_aid101 = 
prove( ` {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)} SUBSET {z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} 
 /\ (CARD ({z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
 <= CARD {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
 --csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}) /\ (FINITE ({z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) ) 
 ==> {z | ~(z = Cx (&0))
 /\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} = 
 {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}  `,
MESON_TAC[CARD_SUBSET; CARD_SUBSET_EQ; LE_ANTISYM]);;


let thm_aid102 = prove(` ~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> CARD {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}  = 2 `,
SIMP_TAC[CARD_CLAUSES; FINITE_INSERT; FINITE_EMPTY; NOT_IN_EMPTY; ARITH]
THEN SIMP_TAC [ IN_SING ] THEN CONV_TAC COMPLEX_FIELD);;


let thm_aid103 = prove(` ~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> (CARD ({z|  (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)})
 <= CARD {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}) `,
SIMP_TAC [goal_thm98;thm_aid102]);;


let coherent_card_set_poles = 
prove( ` ~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> (CARD ({z | ~(z = Cx (&0))
/\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)}) 
<= CARD {csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}) `,
STRIP_TAC THEN MP_TAC thm_aid103 THEN MP_TAC thm_aid99 THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN ASM_ARITH_TAC);;


let coherent_set_poles = 
prove( `  ~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3) = Cx(&0)) 
==> {z | ~(z = Cx (&0))
/\ (z pow 2 +(Cx (sqrt (k1 * k2))*G1*G3 - Cx (sqrt ((&1 - k1) * (&1 - k2)))*G2*G3)) = Cx(&0)} = 
{csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3),
	--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)}  `,
STRIP_TAC THEN MP_TAC coherent_card_set_poles THEN MP_TAC coherent_finite_set_poles
THEN MP_TAC aid_thm96 THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [thm_aid101]);;


let coherent_aid2 = 
	prove ( `norm (--csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - 
	Cx(sqrt(k1*k2))*G1*G3)) = 
	norm (csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - 
	Cx(sqrt(k1*k2))*G1*G3))`, 
	REWRITE_TAC [norm_neg]);;

let dcdr_coherent_stability = 
prove( `!(k1:real) (k2:real) (G1:complex) (G2:complex) (G3:complex).
	((&0<=k1)/\(&0<=k2)/\(&0<=(&1-k1))/\(&0<=(&1-k2))) /\
 	(norm (csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - 
       Cx(sqrt(k1*k2))*G1*G3)) <= &1) /\
	~(csqrt (Cx(sqrt((&1-k1)*(&1-k2)))*G2*G3 - Cx(sqrt(k1*k2))*G1*G3)= Cx(&0)) 
	==> is_stable_psp (\z. DCDR_coherent (G1/z) (G2/z) 
		(G3/z) k1 k2) z `,
REWRITE_TAC [is_stable_psp;poles_in_unit_circle;
	are_valid_poles;in_unit_circle;
	is_valid_pole;set_of_poles] THEN
ONCE_REWRITE_TAC [aid_thm51] THEN
SIMP_TAC [DENOMINATOR_OF_COHERENT_DCDR] THEN
ONCE_REWRITE_TAC [GSYM aid_thm51] THEN
SIMP_TAC [COMPLEX_DIV_EQ_0] THEN
SIMP_TAC [ (COMPLEX_FIELD ` (z pow 2 = Cx(&0)) <=> (z = Cx(&0))`)]
THEN ONCE_REWRITE_TAC [aid_thm61] THEN
SIMP_TAC [coherent_set_poles] THEN
SIMP_TAC [aid_thm50] THEN
REPEAT STRIP_TAC THEN
ASM_SIMP_TAC [] THEN
REPEAT (POP_ASSUM MP_TAC) THEN
SIMP_TAC[ coherent_aid2] THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC[]);;


(*-------------------------------------------------------------------------*)
(*          Resonance conditions of coherent DCDR circuit                  *)
(*-------------------------------------------------------------------------*)


	
let NUMERATOR_OF_COHERENT_DCDR = 
prove( `~(z = Cx (&0)) /\ &0 <= k1 /\ &0 <= k2 /\ &0 <= &1 - k1 /\ &0 <= &1 - k2
 ==>  NUMERATOR (DCDR_coherent (G1/z) (G2/z) (G3/z) k1 k2) =   
 ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) /(z pow 3)`,
SIMP_TAC[NUM_OF_COHERENT_DCDR] THEN CONV_TAC COMPLEX_FIELD);;


let coherent_num_poly = prove(` ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = 
vsum (0..2) (\i. (EL i [
 -- G1*G2*G3;Cx(&0) ; (Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)]) * 
  ( z) pow i)`,
REWRITE_TAC[ONE;TWO;VSUM_22;EL;HD;TL;complex_pow] THEN
CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD);;



let aid_thm111 = prove( ` ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0))
 ==> ~( EL 2
 [-- G1*G2*G3;Cx(&0) ; (Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)]= Cx(&0) )`, 
 SIMP_TAC [EL;HD;TL;ONE;TWO] THEN CONV_TAC NUM_REDUCE_CONV THEN CONV_TAC COMPLEX_FIELD );;


 let thm_aid112 = prove(`~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) ==> ~(2 IN 0..2
==> EL 2
[-- G1*G2*G3;Cx(&0) ; (Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)] = Cx (&0))`,
SIMP_TAC [aid_thm111;thm_aid24]);;


let thm_aid113 = prove( `~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) 
==> ~(!i. i IN 0..2 ==>  (EL i
 [-- G1*G2*G3;Cx(&0) ; (Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)]) = Cx (&0))`,
 REWRITE_TAC[ NOT_FORALL_THM] THEN MESON_TAC [thm_aid112]);;


let thm_aid114 = prove( `~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0))
 ==> (csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) IN {z| ~(z = Cx (&0)) 
 /\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) 
 /\ (--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) IN {z| ~(z = Cx (&0)) 
 /\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)})   `,
REWRITE_TAC[IN_ELIM_THM] THEN REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT]
THEN SIMP_TAC [COMPLEX_POW_2] THEN SIMP_TAC [COMPLEX_MUL_LNEG;COMPLEX_MUL_RNEG]
THEN SIMP_TAC [COMPLEX_NEG_NEG] THEN SIMP_TAC [GSYM COMPLEX_POW_2] THEN
REWRITE_TAC[GSYM COMPLEX_POW_2;CSQRT] THEN
SIMP_TAC [ (COMPLEX_FIELD ` ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) =
   Cx (&0)) ==> ~(--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) =
   Cx (&0))`)] THEN
REWRITE_TAC[CSQRT_EQ_0;COMPLEX_DIV_EQ_0;DE_MORGAN_THM] THEN CONV_TAC COMPLEX_FIELD);;


let thm_aid115 = prove(`~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
==> ( csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) IN {z| ~(z = Cx (&0)) /\ 
((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) /\ 
(--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) IN {z| ~(z = Cx (&0)) /\ 
((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) ==> 
{csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))} SUBSET {z | ~(z = Cx (&0))
 /\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} `,
SET_TAC [ ]);;


let aid_thm116 = prove( `~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
==> {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))} SUBSET {z | ~(z = Cx (&0))
 /\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} `,
STRIP_TAC THEN MP_TAC thm_aid114 THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [thm_aid115]);;


let thm_aid117 = prove( `{z | ~(z = Cx (&0))
/\  ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} 
SUBSET {z |  ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}`,
SIMP_TAC [SUBSET] THEN SIMP_TAC [IN_ELIM_THM]);;


let goal_thm118 = prove(`~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) 
==> CARD {z| ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} <= 2 /\ 
FINITE {z| ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}`,
MP_TAC thm_aid113 THEN SIMP_TAC [coherent_num_poly] THEN SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND]);;


let thm_aid119 = prove(` ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) 
==>  CARD ({z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) 
<= CARD {z | ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}`,
STRIP_TAC THEN MP_TAC thm_aid117 THEN MP_TAC goal_thm118 THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [CARD_SUBSET  ]);;


let coherent_finite_set_zeros = 
prove(` ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) 
==>  FINITE {z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}`,
 STRIP_TAC THEN MP_TAC thm_aid117 THEN MP_TAC goal_thm118 THEN
REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN MESON_TAC [ FINITE_SUBSET]);;


let thm_aid121 = prove( `  {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
	--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))} 
SUBSET {z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} 
/\ (CARD ({z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) 
<= CARD {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
	--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}) 
/\ (FINITE ({z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) ) 
==> {z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} = 
{csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}  `,
MESON_TAC[CARD_SUBSET; CARD_SUBSET_EQ; LE_ANTISYM]);;




let thm_aid122 = 
prove( ` ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
==> CARD {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}  = 2 `,
SIMP_TAC[CARD_CLAUSES; FINITE_INSERT; FINITE_EMPTY; NOT_IN_EMPTY; ARITH]
THEN SIMP_TAC [ IN_SING ] THEN CONV_TAC COMPLEX_FIELD);;



let thm_aid123 = 
prove( ` ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
/\ ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) ==> 
(CARD ({z|  ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) 
<= CARD {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
	--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}) `,
 SIMP_TAC [goal_thm118;thm_aid122]);;


let coherent_card_set_zeros = 
prove(` ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
/\ ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) 
==> (CARD ({z | ~(z = Cx (&0))
/\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)}) 
<= CARD {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
	--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}) `,
STRIP_TAC THEN MP_TAC thm_aid123 THEN MP_TAC thm_aid119
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] THEN ASM_ARITH_TAC);;






let thm_aid125 = 
prove(` ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
==>  ~((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2) = Cx(&0)) `,
SIMP_TAC[CSQRT_EQ_0;COMPLEX_DIV_EQ_0;DE_MORGAN_THM]);;





let coherent_set_zeros = 
prove( `~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0))  
==> {z | ~(z = Cx (&0))
 /\ ((Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)*z pow 2  - G1*G2*G3) = Cx(&0)} = 
 {csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)),
--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2))}  `,
STRIP_TAC THEN MP_TAC thm_aid125 THEN
MP_TAC coherent_card_set_zeros THEN MP_TAC coherent_finite_set_zeros
THEN MP_TAC aid_thm116 THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC [thm_aid121]);;


let coherent_zero_norm = 
	prove ( `norm 
	(--csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - 
	Cx(sqrt(k1*k2))*G2))) = 
	norm (csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - 
	Cx(sqrt(k1*k2))*G2)))`, 
	REWRITE_TAC [norm_neg]);;


let dcdr_coherent_resonance = 
prove( `!(k1:real) (k2:real) (G1:complex) (G2:complex) (G3:complex).
	((&0<=k1)/\(&0<=k2)/\(&0<=(&1-k1))/\(&0<=(&1-k2))) /\ (norm 
	(csqrt ((G1*G2*G3)/
	(Cx(sqrt((&1-k1)*(&1-k2)))*G1- Cx(sqrt(k1*k2))*G2))) <= &1) 
	/\ 
 ~(csqrt ((G1*G2*G3)/(Cx(sqrt((&1-k1)*(&1-k2)))*G1 - Cx(sqrt(k1*k2))*G2)) = Cx(&0)) 
	==> is_resonant_psp (\z. DCDR_coherent (G1/z) (G2/z) 
		(G3/z) k1 k2) z `,
REWRITE_TAC [is_resonant_psp;poles_in_unit_circle;
	are_valid_zeros;in_unit_circle;
	is_valid_zero;set_of_zeros] THEN
ONCE_REWRITE_TAC [aid_thm51] THEN
SIMP_TAC [NUMERATOR_OF_COHERENT_DCDR] THEN
ONCE_REWRITE_TAC [GSYM aid_thm51] THEN
SIMP_TAC [COMPLEX_DIV_EQ_0] THEN
SIMP_TAC [ (COMPLEX_FIELD ` (z pow 3 = Cx(&0)) <=> (z = Cx(&0))`)]
THEN ONCE_REWRITE_TAC [aid_thm61]
THEN SIMP_TAC [coherent_set_zeros]
THEN SIMP_TAC [aid_thm50]
THEN REPEAT STRIP_TAC
THEN ASM_SIMP_TAC []
THEN REPEAT (POP_ASSUM MP_TAC)
THEN SIMP_TAC[coherent_zero_norm]
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC[]);;


(*-----------------------------------------------------------*)
(*   Test of the values of poles given by the book           *)
(*------------------------------------------------------------*)

let fausse_example_help =
prove( `(!x. x = --Cx (&990959 / &1000000) \/ x = Cx (&990959 / &1000000)
          ==> ~(x = Cx (&0)) /\
              Cx (&1) -
              Cx (&9 / &10) *
              Cx (&9 / &10) *
              (Cx (&12 / &10) * inv x) *
              Cx (&1) *
              inv x -
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) * inv x) *
              Cx (&1) *
              inv x =
              Cx (&0)) <=> ( ~((Cx (&990959 / &1000000)) = Cx (&0)) /\
              Cx (&1) -
              Cx (&9 / &10) *
              Cx (&9 / &10) *
              (Cx (&12 / &10) * inv (Cx (&990959 / &1000000)) ) *
              Cx (&1) *
              inv (Cx (&990959 / &1000000)) -
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) * inv (Cx (&990959 / &1000000))) *
              Cx (&1) *
              inv (Cx (&990959 / &1000000)) =
              Cx (&0)) /\ (~((--Cx (&990959 / &1000000)) = Cx (&0)) /\
              Cx (&1) -
              Cx (&9 / &10) *
              Cx (&9 / &10) *
              (Cx (&12 / &10) * inv (--Cx (&990959 / &1000000)) )*
              Cx (&1) *
              inv (--Cx (&990959 / &1000000)) -
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) - Cx (&9 / &10)) *
              (Cx (&1) * inv (--Cx (&990959 / &1000000))) *
              Cx (&1) *
              inv (--Cx (&990959 / &1000000)) =
              Cx (&0))`, MESON_TAC[]);;


let fausse_example = prove(`unstable_psp_given_poles (\z. 
DCDR ((Cx (&12 / &10))*(inv z)) ((Cx (&1))*(inv z)) ((Cx (&1))*(inv z)) 
(Cx (&9 / &10)) (Cx (&9 / &10))) z {Cx (&990959 / &1000000),
	--Cx (&990959 / &1000000)}`,
REWRITE_TAC[unstable_psp_given_poles;is_stable_psp_given_poles;are_valid_poles;
is_valid_pole;poles_in_unit_circle;
 set_of_poles;in_unit_circle;DENOM_OF_DCDR;IN_ELIM_THM;COMPLEX_NORM_CX]
THEN SIMP_TAC [aid_thm50;IN_ELIM_THM;SUBSET] THEN SIMP_TAC [fausse_example_help]
THEN REWRITE_TAC[GSYM complex_div;GSYM CX_SUB;GSYM CX_ADD;
GSYM CX_MUL;GSYM CX_DIV;CX_INJ] 
THEN TRY (CONV_TAC REAL_FIELD));;


let fausse_example2_help =
prove( `(!x. x = --Cx (&905539/ &1000000) \/ x = Cx (&905539 / &1000000)
          ==> ~(x = Cx (&0)) /\
              Cx (&1) -
            Cx (&9 / &10) *
            Cx (&9 / &10) *
            (Cx (&1) * inv x) *
            Cx (&1) *
            inv x -
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) * inv x) *
            Cx (&1) *
            inv x =
              Cx (&0)) <=> ( ~((Cx (&905539 / &1000000)) = Cx (&0)) /\
              Cx (&1) -
            Cx (&9 / &10) *
            Cx (&9 / &10) *
            (Cx (&1) * inv (Cx (&905539 / &1000000))) *
            Cx (&1) *
            inv (Cx (&905539 / &1000000)) -
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) * inv (Cx (&905539 / &1000000))) *
            Cx (&1) *
            inv (Cx (&905539 / &1000000)) =
              Cx (&0)) /\ (~((--Cx (&905539 / &1000000)) = Cx (&0)) /\
              Cx (&1) -
            Cx (&9 / &10) *
            Cx (&9 / &10) *
            (Cx (&1) * inv (--Cx (&905539 / &1000000))) *
            Cx (&1) *
            inv (--Cx (&905539 / &1000000)) -
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) - Cx (&9 / &10)) *
            (Cx (&1) * inv (--Cx (&905539 / &1000000))) *
            Cx (&1) *
            inv (--Cx (&905539 / &1000000)) =
              Cx (&0))`, MESON_TAC[]);;	
		
let fausse_example2 = prove(`unstable_psp_given_poles (\z. 
DCDR ((Cx (&1))*(inv z)) ((Cx (&1))*(inv z)) ((Cx (&1))*(inv z)) 
	      (Cx (&9 / &10)) (Cx (&9 / &10))) z {Cx (&905539 / &1000000),
	      --Cx (&905539 / &1000000)}`,
REWRITE_TAC[unstable_psp_given_poles;is_stable_psp_given_poles;are_valid_poles;
is_valid_pole;poles_in_unit_circle;
set_of_poles;in_unit_circle;DENOM_OF_DCDR;IN_ELIM_THM;COMPLEX_NORM_CX]
THEN SIMP_TAC [aid_thm50;IN_ELIM_THM;SUBSET] THEN SIMP_TAC [fausse_example2_help]
THEN REWRITE_TAC[GSYM complex_div;GSYM CX_SUB;GSYM CX_ADD;
GSYM CX_MUL;GSYM CX_DIV;CX_INJ] 
THEN TRY (CONV_TAC REAL_FIELD));;
(*--------------------------------------------------------------------*)
(*                       ADD/DROP Filter                              *)
(*--------------------------------------------------------------------*)

let aid0 = prove ( `--(ii * s1) * 
	csqrt (cexp ((--a * L) / Cx (&2)) * inv z) * --(ii * s2) = 
	-- s1*s2*csqrt (cexp ((--a * L) / Cx (&2)) * inv z)`, 
	CONV_TAC COMPLEX_FIELD);;
	
let aid1 = prove ( `c1 *
  csqrt (cexp ((--a * L) / Cx (&2)) * inv z) *
  c2 *
  csqrt (cexp ((--a * L) / Cx (&2)) * inv z) = 
  c1*c2*csqrt (cexp ((--a * L) / Cx (&2)) * inv z) *
  csqrt (cexp ((--a * L) / Cx (&2)) * inv z)`, 
  CONV_TAC COMPLEX_FIELD);;
  
let aid2 = prove ( `c1 *
  csqrt (cexp ((--a * L) / Cx (&2)) * inv z) *
  c2 *
  csqrt (cexp ((--a * L) / Cx (&2)) * inv z) = 
  c1*c2*(cexp ((--a * L) / Cx (&2)) * inv z)`, 
  REWRITE_TAC [aid1;GSYM COMPLEX_POW_2;CSQRT]);;

let aid3 = prove ( `  
	(--a * L) / Cx (&2) = --a * L / Cx (&2)`, 
	CONV_TAC COMPLEX_FIELD);;
	

let ADD_DROP_FILTER = prove (` ~(z=Cx(&0)) ==>
 SFG_Main ([1,--(ii*s1),4;1,c1,3;2,c1,4;2,--(ii*s1),3;4,csqrt ((cexp ((--a*L)/(Cx (&2)))) * inv z)
,5;5,c2,7;
5,--(ii*s2),8;6,c2,8;6,--(ii*s2),7;7,csqrt ((cexp ((--a*L)/(Cx
(&2)))) * inv z),2], 8, 8,1) =
((--s1 * s2 * (csqrt ((cexp (--a*L/(Cx (&2)))) * inv z)))
/ (Cx(&1) - c1 * c2 * ((cexp (--a*L/(Cx (&2)))) * inv z)))`,
SFG_TAC
THEN SIMP_TAC [aid0;aid2; COMPLEX_ADD_RID;COMPLEX_ADD_LID;
	COMPLEX_MUL_RID;COMPLEX_MUL_LID;CEXP_COMPLEX;aid3 ]
THEN COMPLEX_DIV_TAC);;



(*----------------------------------------------------------------*) 
(*--------------------------Mulit legged Robot--------------------*)  
(*----------------------------------------------------------------*) 
 
let Legged_Robot = 
	prove (`! (G1:complex) G2 G3 G4 G5 G6 G7 G8 H2 H3  H6 H7.  
   SFG_Main ([1,G1,2;1,G5,5;2,G2,3;3,H2,2;3,G3,4;4,H3,3;4,G4,8;5,G6,6; 
   6,H6,5;6,G7,7;7,H7,6;7,G8,8], 8, 8,1) =  
(G1*G2*G3*G4*(Cx(&1) - G6*H6 - G7*H7) +  
G5*G6*G7*G8*(Cx(&1) - G2*H2 - G3*H3)) / 
(Cx(&1) - G2*H2 - G3*H3 - G6*H6 - G7*H7 + G2*H2*G6*H6 + G2*H2*G7*H7 +  
 G3*H3*G6*H6 + G3*H3*G7*H7)`,
SFG_TAC
THEN COMPLEX_DIV_TAC);; 
