(* ========================================================================= *)
(*                                                                           *)
(*                        Quadruple.ml                                   *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2015                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <beillahi@encs.concordia.ca>,                                *)
(*       	<muh_sidd@ece.concordia.ca>                                 *)
(*                                                                           *)
(* Last update: January 26, 2015                                               *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)


needs "Mason_Formula.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)


let Quadruple_TF = prove(` ~(z=Cx(&0)) ==> 
SFG_Main ([1,c1,3;1,--ii*s1,4;2,c1,4;
2,--ii*s1,3; 4,csqrt ((inv z) pow k1) ,5;5,c2,7 ;5,--ii*s2,8 
;6,c2,8;6,--ii*s2,7;7,csqrt ((inv z) pow k1) ,2 ;
8,csqrt ((inv z) pow k2) ,9;9,c3,11;9,--ii*s3,12 ;10,c3,12;
10,--ii*s3,11;11,csqrt ((inv z) pow k2) ,6 ;
12,csqrt ((inv z) pow k3) ,13;13,c4,15; 13,--ii*s4,16 ;
14,c4,16 ;14,--ii*s4,15;15,csqrt ((inv z) pow k3) ,10 ;
16,csqrt ((inv z) pow k4) ,17;17,c5,19;17,--ii*s5,20;18,c5,20;
18,--ii*s5,19;19,csqrt ((inv z) pow k4) ,14], 20 ,20, 1) = 
   (((--ii * s1) *
       csqrt (inv z pow k1) *
       (--ii * s2) *
       csqrt (inv z pow k2) *
       (--ii * s3) *
       csqrt (inv z pow k3) *
       (--ii * s4) *
       csqrt (inv z pow k4) *
       --ii *
       s5) *
      Cx (&1)) /
     (Cx (&1) +
      ((c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c2 *
       csqrt (inv z pow k2) *
       c3 *
       csqrt (inv z pow k2) +
       (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c2 *
       csqrt (inv z pow k2) *
       (--ii * s3) *
       csqrt (inv z pow k3) *
       c4 *
       csqrt (inv z pow k3) *
       (--ii * s3) *
       csqrt (inv z pow k2) +
       (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c2 *
       csqrt (inv z pow k2) *
       (--ii * s3) *
       csqrt (inv z pow k3) *
       (--ii * s4) *
       csqrt (inv z pow k4) *
       c5 *
       csqrt (inv z pow k4) *
       (--ii * s4) *
       csqrt (inv z pow k3) *
       (--ii * s3) *
       csqrt (inv z pow k2) +
       (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c3 *
       csqrt (inv z pow k3) *
       c4 *
       csqrt (inv z pow k3) +
       (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c3 *
       csqrt (inv z pow k3) *
       (--ii * s4) *
       csqrt (inv z pow k4) *
       c5 *
       csqrt (inv z pow k4) *
       (--ii * s4) *
       csqrt (inv z pow k3) +
       (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
       c4 *
       csqrt (inv z pow k4) *
       c5 *
       csqrt (inv z pow k4)) +
      --((c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
         (c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) *
         c4 *
         csqrt (inv z pow k4) *
         c5 *
         csqrt (inv z pow k4)) -
      (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
      (c2 *
       csqrt (inv z pow k2) *
       (--ii * s3) *
       csqrt (inv z pow k3) *
       c4 *
       csqrt (inv z pow k3) *
       (--ii * s3) *
       csqrt (inv z pow k2)) *
      c4 *
      csqrt (inv z pow k4) *
      c5 *
      csqrt (inv z pow k4) -
      (c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) *
      (((c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
        c3 *
        csqrt (inv z pow k3) *
        c4 *
        csqrt (inv z pow k3) +
        (c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
        c3 *
        csqrt (inv z pow k3) *
        (--ii * s4) *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4) *
        (--ii * s4) *
        csqrt (inv z pow k3) +
        (c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
        c4 *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4)) +
       --((c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
          (c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) *
          c4 *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4))) +
      (((c1 *
         csqrt (inv z pow k1) *
         (--ii * s2) *
         csqrt (inv z pow k2) *
         c3 *
         csqrt (inv z pow k2) *
         (--ii * s2) *
         csqrt (inv z pow k1)) *
        c3 *
        csqrt (inv z pow k3) *
        c4 *
        csqrt (inv z pow k3) +
        (c1 *
         csqrt (inv z pow k1) *
         (--ii * s2) *
         csqrt (inv z pow k2) *
         c3 *
         csqrt (inv z pow k2) *
         (--ii * s2) *
         csqrt (inv z pow k1)) *
        c3 *
        csqrt (inv z pow k3) *
        (--ii * s4) *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4) *
        (--ii * s4) *
        csqrt (inv z pow k3) +
        (c1 *
         csqrt (inv z pow k1) *
         (--ii * s2) *
         csqrt (inv z pow k2) *
         c3 *
         csqrt (inv z pow k2) *
         (--ii * s2) *
         csqrt (inv z pow k1)) *
        c4 *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4)) +
       --((c1 *
           csqrt (inv z pow k1) *
           (--ii * s2) *
           csqrt (inv z pow k2) *
           c3 *
           csqrt (inv z pow k2) *
           (--ii * s2) *
           csqrt (inv z pow k1)) *
          (c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) *
          c4 *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4)) +
       ((c1 *
         csqrt (inv z pow k1) *
         (--ii * s2) *
         csqrt (inv z pow k2) *
         (--ii * s3) *
         csqrt (inv z pow k3) *
         c4 *
         csqrt (inv z pow k3) *
         (--ii * s3) *
         csqrt (inv z pow k2) *
         (--ii * s2) *
         csqrt (inv z pow k1)) *
        c4 *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4) +
        (((c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
          c3 *
          csqrt (inv z pow k3) *
          c4 *
          csqrt (inv z pow k3) +
          (c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
          c3 *
          csqrt (inv z pow k3) *
          (--ii * s4) *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4) *
          (--ii * s4) *
          csqrt (inv z pow k3) +
          (c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
          c4 *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4)) +
         --((c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) *
            (c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) *
            c4 *
            csqrt (inv z pow k4) *
            c5 *
            csqrt (inv z pow k4)) +
         ((c2 *
           csqrt (inv z pow k2) *
           (--ii * s3) *
           csqrt (inv z pow k3) *
           c4 *
           csqrt (inv z pow k3) *
           (--ii * s3) *
           csqrt (inv z pow k2)) *
          c4 *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4) +
          ((c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) *
           c4 *
           csqrt (inv z pow k4) *
           c5 *
           csqrt (inv z pow k4) +
           --(c4 * csqrt (inv z pow k4) * c5 * csqrt (inv z pow k4)) -
           c3 *
           csqrt (inv z pow k3) *
           (--ii * s4) *
           csqrt (inv z pow k4) *
           c5 *
           csqrt (inv z pow k4) *
           (--ii * s4) *
           csqrt (inv z pow k3) -
           c3 * csqrt (inv z pow k3) * c4 * csqrt (inv z pow k3)) -
          c2 *
          csqrt (inv z pow k2) *
          (--ii * s3) *
          csqrt (inv z pow k3) *
          (--ii * s4) *
          csqrt (inv z pow k4) *
          c5 *
          csqrt (inv z pow k4) *
          (--ii * s4) *
          csqrt (inv z pow k3) *
          (--ii * s3) *
          csqrt (inv z pow k2) -
          c2 *
          csqrt (inv z pow k2) *
          (--ii * s3) *
          csqrt (inv z pow k3) *
          c4 *
          csqrt (inv z pow k3) *
          (--ii * s3) *
          csqrt (inv z pow k2)) -
         c2 * csqrt (inv z pow k2) * c3 * csqrt (inv z pow k2)) -
        c1 *
        csqrt (inv z pow k1) *
        (--ii * s2) *
        csqrt (inv z pow k2) *
        (--ii * s3) *
        csqrt (inv z pow k3) *
        (--ii * s4) *
        csqrt (inv z pow k4) *
        c5 *
        csqrt (inv z pow k4) *
        (--ii * s4) *
        csqrt (inv z pow k3) *
        (--ii * s3) *
        csqrt (inv z pow k2) *
        (--ii * s2) *
        csqrt (inv z pow k1) -
        c1 *
        csqrt (inv z pow k1) *
        (--ii * s2) *
        csqrt (inv z pow k2) *
        (--ii * s3) *
        csqrt (inv z pow k3) *
        c4 *
        csqrt (inv z pow k3) *
        (--ii * s3) *
        csqrt (inv z pow k2) *
        (--ii * s2) *
        csqrt (inv z pow k1)) -
       c1 *
       csqrt (inv z pow k1) *
       (--ii * s2) *
       csqrt (inv z pow k2) *
       c3 *
       csqrt (inv z pow k2) *
       (--ii * s2) *
       csqrt (inv z pow k1)) -
      c1 * csqrt (inv z pow k1) * c2 * csqrt (inv z pow k1)) `, 
	  SFG_TAC THEN SIMP_TAC [COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO;
	  COMPLEX_ADD_LID;COMPLEX_ADD_RID;COMPLEX_SUB_LZERO;
	  COMPLEX_SUB_RZERO;COMPLEX_NEG_0]);;
